"""
Flask commands to manage torrents on Put.IO.
"""
import os
import pyinotify

from flask_script import Manager
from putio_automator.db import with_db
from putio_automator.manage import app


manager = Manager(usage='Manage torrents')

@manager.command
def add(parent_id=0):
    "Add a torrent"
    files = os.listdir(app.config['TORRENTS'])

    if len(files):
        def func(connection):
            "Anonymous function"
            conn = connection.cursor()

            for name in os.listdir(app.config['TORRENTS']):
                path = os.path.join(app.config['TORRENTS'], name)
                size = os.path.getsize(path)

                conn.execute("select datetime(created_at, 'localtime') from torrents where name = ? and size = ?", (name, size))
                row = conn.fetchone()

                if row is None:
                    try:
                        app.logger.debug('adding torrent: %s' % path)
                        transfer = app.client.Transfer.add_torrent(path, parent_id=parent_id)
                        os.unlink(path)
                        app.logger.info('added transfer: %s' % transfer)
                    except Exception as e:
                        if e.message == 'BadRequest':
                            # Assume it's already added
                            os.unlink(path)
                            app.logger.warning('deleted torrent, already added : %s' % (name,))
                        else:
                            raise e

                    conn.execute('insert into torrents (name, size) values (?, ?)', (name, size))
                    connection.commit()
                else:
                    os.unlink(path)
                    app.logger.warning('deleted torrent, added at %s : %s' % (row[0], name))

        with_db(app, func)

@manager.command
def watch(parent_id=0):
    "Watch a folder for new torrents to add"
    add()

    class EventHandler(pyinotify.ProcessEvent):
        "Event handler for responding to a new or updated torrent file"
        def process_IN_CLOSE_WRITE(self, event):
            "Do the above"
            app.logger.debug('adding torrent, received event: %s' % event)
            transfer = app.client.Transfer.add_torrent(event.pathname, parent_id=parent_id)
            os.unlink(event.pathname)
            app.logger.info('added transfer: %s' % transfer)

    watch_manager = pyinotify.WatchManager()
    mask = pyinotify.IN_CLOSE_WRITE

    handler = EventHandler()
    notifier = pyinotify.Notifier(watch_manager, handler)

    wdd = watch_manager.add_watch(app.config['TORRENTS'], mask, rec=True)
    app.logger.debug('added watch: %s' % wdd)

    notifier.loop()



"""
Initialize the application.
"""
import datetime
import distutils.dir_util
import os
import sqlite3

try:
    import putiopy as putio
except ImportError:
    import putio

from flask import Flask

import appdirs

APP_NAME = 'putio-automator'
APP_AUTHOR = 'datashaman'
APP_TAG = '%s/%s' % (APP_AUTHOR, APP_NAME)

from db import create_db, database_path

DIRS = appdirs.AppDirs(APP_NAME, APP_AUTHOR)

def create_app(config=None):
    "Create a Flask app given config"
    app = Flask(__name__)

    distutils.dir_util.mkpath(DIRS.user_data_dir)

    if config is None:
        config = find_config()

    if config is not None:
        app.config.from_pyfile(config)

    create_db(app)
    app.client = create_client(app)

    return app

def create_client(app):
    "Create a Put.IO client"
    if 'PUTIO_TOKEN' in app.config:
        client = putio.Client(app.config['PUTIO_TOKEN'], use_retry=True)

        return client

def date_handler(obj):
    "Date handler for JSON serialization"
    if isinstance(obj, datetime.datetime) or isinstance(obj, datetime.date):
        return obj.isoformat()
    else:
        return None

def find_config():
    "Search for config on wellknown paths"
    search_paths = [
        os.path.join(os.getcwd(), 'config.py'),
        os.path.join(DIRS.user_data_dir, 'config.py'),
        os.path.join(DIRS.site_data_dir, 'config.py'),
    ]

    config = None

    for search_path in search_paths:
        if os.path.exists(search_path) and not os.path.isdir(search_path):
            config = search_path
            break

    return config



"""
Flask command for managing files on Put.IO.
"""
import json
import os
import shutil
 
from flask_script import Command, Manager
from putio_automator import date_handler
from putio_automator.db import with_db
from putio_automator.manage import app


manager = Manager(usage='Manage files')

class List(Command):
    "List files: Create a Flask command manually - clashes with list"
    def run(self, parent_id=0):
        "Run the command"
        files = app.client.File.list(parent_id)
        print json.dumps([vars(f) for f in files], indent=4, default=date_handler)

manager.add_command('list', List())

@manager.command
def download(limit=None, chunk_size=256, parent_id=0):
    "Download files"
    files = app.client.File.list(parent_id)
    app.logger.info('%s files found' % len(files))

    if len(files):
        def func(connection):
            "Anonymous func"
            if limit is not None:
                downloaded = 0

            conn = connection.cursor()

            for current_file in files:
                conn.execute("select datetime(created_at, 'localtime') from downloads where name = ? and size = ?", (current_file.name, current_file.size))
                row = conn.fetchone()

                if row is None:
                    app.logger.debug('downloading file: %s' % current_file)
                    current_file.download(dest=app.config['INCOMPLETE'], delete_after_download=True, chunk_size=int(chunk_size)*1024)
                    app.logger.info('downloaded file: %s' % current_file)

                    path = os.path.join(app.config['INCOMPLETE'], current_file.name)
                    shutil.move(path, app.config['DOWNLOADS'])

                    conn.execute('insert into downloads (id, name, size) values (?, ?, ?)', (current_file.id, current_file.name, current_file.size))
                    connection.commit()

                    if limit is not None:
                        downloaded = downloaded + 1

                        if downloaded > limit:
                            break
                else:
                    app.logger.warning('file already downloaded at %s : %s' % (row[0], current_file))

        with_db(app, func)



"""
Flask commands
"""
from putio_automator.commands.account import manager as account
from putio_automator.commands.config import manager as config
from putio_automator.commands.db import manager as db
from putio_automator.commands.docker import manager as docker
from putio_automator.commands.files import manager as files
from putio_automator.commands.torrents import manager as torrents
from putio_automator.commands.transfers import manager as transfers



"""
Flask commands for managing Put.IO account
"""

import json

from flask_script import Manager
from putio_automator import date_handler
from putio_automator.manage import app

manager = Manager(usage='Manage account')


@manager.command
def info():
    "Show account info"
    print json.dumps(app.client.Account.info(), indent=4, default=date_handler)



#!/usr/bin/env python
"""
Flask script aka CLI entry module.
"""

import distutils.dir_util
import logging
import os

from flask_script import Manager, prompt
from logging.handlers import RotatingFileHandler
from putio_automator import create_app, APP_NAME, APP_AUTHOR

import appdirs

app = create_app()

manager = Manager(app, usage='Manage torrents and downloads on Put.io')

# This must be here (circular references)
import commands

manager.add_command('account', commands.account)
manager.add_command('config', commands.config)
manager.add_command('db', commands.db)
manager.add_command('docker', commands.docker)
manager.add_command('files', commands.files)
manager.add_command('torrents', commands.torrents)
manager.add_command('transfers', commands.transfers)

def main():
    "Main entry point"
    log_dir = os.getenv('LOG_DIR', appdirs.user_log_dir(APP_NAME, APP_AUTHOR))
    distutils.dir_util.mkpath(log_dir)

    logfile_path = os.path.join(log_dir, 'application.log')

    handler = RotatingFileHandler(logfile_path, maxBytes=10000000, backupCount=5)

    formatter = logging.Formatter("[%(asctime)s] {%(pathname)s:%(lineno)d} %(levelname)s - %(message)s")
    handler.setFormatter(formatter)

    handler.setLevel(app.config.get('LOG_LEVEL', logging.WARNING))

    app.logger.addHandler(handler)

    manager.run()

if __name__ == '__main__':
    main()



"""
Flask commands for managing config of the application
"""

import json
import os
import stat

from flask_script import Manager, prompt
from putio_automator import APP_NAME, APP_AUTHOR, date_handler, find_config
from putio_automator.manage import app

import appdirs


manager = Manager(usage='Manage configuration')

def find_config_dist():
    "Search for the config.py.dist file"
    search_paths = [
        os.path.join(os.getcwd(), 'etc', 'config.py.dist'),
        os.path.join(os.getenv('HOME'), '.local', 'etc', APP_NAME, 'config.py.dist'),
        os.path.join('/etc', APP_NAME, 'config.py.dist'),
    ]

    config = None

    for search_path in search_paths:
        if os.path.exists(search_path) and not os.path.isdir(search_path):
            config = search_path
            break

    return config

@manager.command
def init():
    "Prompt the user for config"
    user_data_dir = appdirs.user_data_dir(APP_NAME, APP_AUTHOR)

    incomplete = os.path.realpath(prompt('Incomplete directory', 'incomplete'))
    downloads = os.path.realpath(prompt('Downloads directory', 'downloads'))
    torrents = os.path.realpath(prompt('Torrents directory', 'torrents'))

    putio_token = prompt('OAuth Token')

    config_path = os.path.realpath(prompt('Config file to write',
                                          os.path.join(user_data_dir,
                                                       'config.py')))

    root = os.getenv('VIRTUAL_ENV')

    if root is None:
        root = '/'

    config_dist = find_config_dist()
    with open(config_dist, 'r') as source:
        contents = (source.read()
                    .replace("os.getenv('PUTIO_TOKEN')",
                             "os.getenv('PUTIO_TOKEN', '" + putio_token + "')")
                    .replace("/files/downloads", downloads)
                    .replace("/files/incomplete", incomplete)
                    .replace("/files/torrents", torrents))

        with open(config_path, 'w') as destination:
            destination.write(contents)

        os.chmod(config_path, stat.S_IRUSR | stat.S_IWUSR)

@manager.command
def show():
    "Show config filename and current config"
    print 'Config filename: %s' % find_config()
    print 'Current config:\n%s' % json.dumps(app.config, indent=4, default=date_handler)



"""
Flask commands to manage the download database
"""
from flask_script import Manager
from putio_automator.db import with_db
from putio_automator.manage import app


manager = Manager(usage='Manage download database')

@manager.command
def forget(name):
    "Delete records of previous downloads in the database"

    def func(connection):
        "Do the above"
        conn = connection.cursor()
        conn.execute('delete from downloads where name like ?', ('%%%s%%' % name,))
        print 'Affected rows: %s' % conn.rowcount

    with_db(app, func)



"""
Flask commands to build and manage a docker instance.
"""
import os
import subprocess

from flask_script import Manager
from putio_automator import APP_TAG
from putio_automator.manage import app


manager = Manager(usage='Manage docker instance')

@manager.command
def pull(tag=APP_TAG):
    "Pull latest application image"
    subprocess.call([
        'docker',
        'pull',
        tag
    ])

@manager.command
def build(tag=APP_TAG):
    "Build an application image"
    subprocess.call([
        'docker',
        'build',
        '-t',
        tag,
        '.'
    ])

@manager.command
@manager.option('-d', '--log-dir', dest='log_dir')
def run(start_hour=0, end_hour=24, check_downloads_every=15, tag=APP_TAG, dir=None, level=None):
    "Run an application container"
    if level is None:
        level = app.config['LOG_LEVEL']

    subprocess.call([
	    'docker',
	    'run',
	    '--rm',
	    '-i',
	    '-e',
	    'START_HOUR=%s' % start_hour,
	    '-e',
	    'END_HOUR=%s' % end_hour,
	    '-e',
	    'CHECK_DOWNLOADS_EVERY=%s' % check_downloads_every,
	    '-e',
	    'LOG_DIR=%s' % dir,
	    '-e',
	    'LOG_LEVEL=%s' % level,
	    '-e',
	    'PUTIO_TOKEN=%s' % app.config['PUTIO_TOKEN'],
	    '-p',
        '9001:9001',
        '-v',
        '%s:/files/incomplete' % app.config['INCOMPLETE'],
        '-v',
        '%s:/files/downloads' % app.config['DOWNLOADS'],
        '-v',
        '%s:/files/torrents' % app.config['TORRENTS'],
        tag
    ])

@manager.command
def bootstrap(start_hour=None, end_hour=None, check_downloads_every=None):
    "Bootstrap the application by cogging a new cron schedule and starting supervisor"
    if start_hour is None:
        start_hour = os.getenv('START_HOUR', 0)

    if end_hour is None:
        end_hour = os.getenv('END_HOUR', 0)

    if check_downloads_every is None:
        check_downloads_every = os.getenv('CHECK_DOWNLOADS_EVERY', 15)

    subprocess.call([
        'cog.py',
        '-r',
        '-D',
        'START_HOUR=%s' % start_hour,
        '-D',
        'END_HOUR=%s' % end_hour,
        '-D',
        'CHECK_DOWNLOADS_EVERY=%s' % check_downloads_every,
        '/etc/cron.d/putio-automator'
    ])

    subprocess.call([
        'supervisord',
        '-n',
        '-c',
        '/etc/supervisor/supervisord.conf',
    ])



"""
Flask commands for managing transfers on Put.IO.
"""
import json

from flask_script import Command, Manager
from putio_automator import date_handler
from putio_automator.manage import app


manager = Manager(usage='Manage transfers')

@manager.command
def cancel_by_status(statuses):
    "Cancel transfers by status"
    if isinstance(statuses, str):
        statuses = statuses.split(',')

    transfer_ids = []
    for transfer in app.client.Transfer.list():
        if transfer.status in statuses:
            transfer_ids.append(transfer.id)

    if len(transfer_ids):
        app.client.Transfer.cancel_multi(transfer_ids)

@manager.command
def cancel_completed():
    "Cancel completed transfers"
    cancel_by_status('COMPLETED')

@manager.command
def cancel_seeding():
    "Cancel seeding transfers"
    cancel_by_status('SEEDING')

@manager.command
def clean():
    "Clean finished transfers"
    app.client.Transfer.clean()

@manager.command
def groom():
    "Cancel seeding and completed transfers, and clean afterwards"
    cancel_by_status(['SEEDING', 'COMPLETED'])
    clean()

class List(Command):
    "List transfers: Manually create Flask command cos of name clash with list"
    def run(self):
        "List transfers"
        transfers = app.client.Transfer.list()
        print json.dumps([vars(t) for t in transfers], indent=4, default=date_handler)
manager.add_command('list', List())
